from uuid import uuid4

from math import floor
import numpy as np
import pandas as pd
from pathlib import Path

from skimage.measure import find_contours, label, regionprops_table
from sklearn.preprocessing import PolynomialFeatures
from sklearn.linear_model import LinearRegression

from model_server.base.accessors import GenericImageDataAccessor, InMemoryDataAccessor, write_accessor_data_to_file
from model_server.base.models import InstanceSegmentationModel
from model_server.base.process import pad, rescale, resample_to_8bit

from model_server.extensions.chaeo.annotators import draw_boxes_on_3d_image
from model_server.extensions.chaeo.products import export_patches_from_zstack, export_multichannel_patches_from_zstack
from extensions.chaeo.params import RoiFilter, RoiSetMetaParams, RoiSetExportParams
from model_server.extensions.chaeo.accessors import MonoPatchStack
from model_server.extensions.chaeo.process import mask_largest_object
from model_server.extensions.chaeo.products import get_patches_from_zmask_meta, get_patch_masks, export_patch_masks


def get_label_ids(acc_seg_mask):
    return label(acc_seg_mask.data[:, :, 0, 0]).astype('uint16')

class RoiSet(object):

    def __init__(
            self,
            acc_obj_ids: GenericImageDataAccessor,  # TODO: enforce subtype of binary or label ID mask
            acc_raw: GenericImageDataAccessor,
            params: RoiSetMetaParams = RoiSetMetaParams(),
    ):
        self.acc_obj_ids = acc_obj_ids
        self.acc_raw = acc_raw
        self.params = params

        self._df = self.filter_df(
            self.make_df(
                self.acc_raw, self.acc_obj_ids, expand_box_by=params.expand_box_by
            ),
            params.filters,
        )

        # temporarily build zmask meta here
        meta = []
        for ob in self.get_df().itertuples(name='LabeledObject'):
            sl = np.s_[ob.ebb_y0: ob.ebb_y1, ob.ebb_x0: ob.ebb_x1, :, ob.ebb_z0: ob.ebb_z1 + 1]  # TODO: on-the-fly in RoiSet, given DF

            # compute contours
            obmask = (acc_obj_ids == ob.label)  # TODO: on-the-fly
            contour = find_contours(obmask)  # TODO: on-the-fly
            mask = obmask[ob.y0: ob.y1, ob.x0: ob.x1]

            rbb = {  # TODO: just put in the DF
                'y0': ob.rel_y0,
                'y1': ob.rel_y1,
                'x0': ob.rel_x0,
                'x1': ob.rel_x1,
            }

            meta.append({
                'df_index': ob.Index,
                'info': ob,
                'slice': sl,
                'relative_bounding_box': rbb,  # TODO: put in DF
                'contour': contour,  # TODO: delegate to getter
                'mask': mask  # TODO: delegate to getter
            })
        self.zmask_meta = meta

        # return intermediate image arrays  # TODO: make on-the-fly
        self.interm = {
            'label_map': acc_obj_ids,
            'argmax': acc_raw.data.argmax(axis=3, keepdims=True)[:, :, 0, 0].astype('uint16'),
        }

        self.count = len(self.zmask_meta)
        self.object_id_labels = self.interm['label_map']
        self.object_class_map = None

    @staticmethod
    def make_df(acc_raw, acc_obj_ids, expand_box_by) -> pd.DataFrame:
        """
        Build dataframe associate object IDs with summary stats
        :param acc_raw: accessor to raw image data
        :param acc_obj_ids: accessor to map of object IDs
        :param expand_box_by: number of pixels to expand bounding box in all directions (without exceeding image boundary)
        :return: pd.DataFrame
        """
        # build dataframe of objects, assign z index to each object
        argmax = acc_raw.data.argmax(axis=3, keepdims=True)[:, :, 0, 0].astype('uint16')
        df = (
            pd.DataFrame(
                regionprops_table(
                    acc_obj_ids,
                    intensity_image=argmax,
                    properties=('label', 'area', 'intensity_mean', 'solidity', 'bbox', 'centroid')
                )
            )
            .rename(
                columns={'bbox-0': 'y0', 'bbox-1': 'x0', 'bbox-2': 'y1', 'bbox-3': 'x1', }
            )
        )
        df['zi'] = df['intensity_mean'].round().astype('int')

        # compute expanded bounding boxes
        h, w, c, nz = acc_raw.shape
        ebxy, ebz = expand_box_by
        df['ebb_y0'] = (df.y0 - ebxy).apply(lambda x: max(x, 0))
        df['ebb_y1'] = (df.y1 + ebxy).apply(lambda x: min(x, h))
        df['ebb_x0'] = (df.x0 - ebxy).apply(lambda x: max(x, 0))
        df['ebb_x1'] = (df.x1 + ebxy).apply(lambda x: min(x, w))
        df['ebb_z0'] = (df.zi - ebz).apply(lambda x: max(x, 0))
        df['ebb_z1'] = (df.zi + ebz).apply(lambda x: min(x, nz))
        df['ebb_h'] = df['ebb_y1'] - df['ebb_y0']
        df['ebb_w'] = df['ebb_x1'] - df['ebb_x0']
        df['ebb_nz'] = df['ebb_z1'] - df['ebb_z0'] + 1

        # compute relative bounding boxes
        df['rel_y0'] = df.y0 - df.ebb_y0
        df['rel_y1'] = df.y1 - df.ebb_y0
        df['rel_x0'] = df.x0 - df.ebb_x0
        df['rel_x1'] = df.x1 - df.ebb_x0

        assert np.all(df['rel_x1'] <= (df['ebb_x1'] - df['ebb_x0']))
        assert np.all(df['rel_y1'] <= (df['ebb_x1'] - df['ebb_x0']))

        df['slice'] = df.apply(
            lambda r:
            np.s_[int(r.ebb_y0): int(r.ebb_y1), int(r.ebb_x0): int(r.ebb_x1), :, int(r.ebb_z0): int(r.ebb_z1) + 1],
            axis=1
        )
        df['relative_slice'] = df.apply(
            lambda r:
            np.s_[int(r.rel_y0): int(r.rel_y1), int(r.rel_x0): int(r.rel_x1), :, :],
            axis=1
        )
        df['mask'] = df.apply(
            lambda r: (acc_obj_ids == r.label)[r.y0: r.y1, r.x0: r.x1],
            axis=1
        )
        return df


    @staticmethod
    def filter_df(df: pd.DataFrame, filters: RoiFilter = None) -> pd.DataFrame:
        query_str = 'label > 0'  # always true
        if filters is not None:  # parse filters
            for k, val in filters.dict(exclude_unset=True).items():
                assert k in ('area', 'solidity')
                vmin = val['min']
                vmax = val['max']
                assert vmin >= 0
                query_str = query_str + f' & {k} > {vmin} & {k} < {vmax}'
        return df.loc[df.query(query_str).index, :]

    def get_df(self) -> pd.DataFrame:  # TODO: exclude columns that refer to objects
        return self._df

    def get_slices(self) -> pd.Series:
        return self.get_df()['slice']

    def add_df_col(self, name, se: pd.Series) -> None:
        self._df[name] = se

    def get_multichannel_projection(self):  # TODO: document and test
        dff = self.df[self.df['keeper']]
        if self.count:
            projected = project_stack_from_focal_points(
                dff['centroid-0'].to_numpy(),
                dff['centroid-1'].to_numpy(),
                dff['zi'].to_numpy(),
                self.acc_raw,
                degree=4,
            )
        else:  # else just return MIP
            projected = self.acc_raw.data.max(axis=-1)
        return projected

    def get_object_mask_by_class(self, class_id):
        return self.object_id_labels == class_id

    def export_patch_masks(self, where, **kwargs) -> list:
        return export_patch_masks(self, where, **kwargs)

    def get_raw_patches(self, channel):
        return get_patches_from_zmask_meta(
            self.acc_raw.get_one_channel_data(channel),
            self.zmask_meta
        )

    def get_zmask(self, mask_type='boxes'):
        """
        Return a mask of same dimensionality as raw data

        :param kwargs: variable-length keyword arguments
            mask_type: if 'boxes', zmask is True in each object's complete bounding box; otherwise 'contours'
        """

        assert mask_type in ('contours', 'boxes')
        zi_st = np.zeros(self.acc_raw.shape, dtype='bool')
        lamap = self.acc_obj_ids

        # make an object map where label is replaced by focus position in stack and background is -1
        lut = np.zeros(lamap.max() + 1) - 1
        df = self.get_df()
        lut[df.label] = df.zi

        if mask_type == 'contours':
            zi_map = (lut[lamap] + 1.0).astype('int')
            idxs = np.array(zi_map) - 1
            np.put_along_axis(
                zi_st,
                np.expand_dims(idxs, (2, 3)),
                1,
                axis=3
            )

            # change background level from to 0 in final frame
            zi_st[:, :, :, -1][lamap == 0] = 0

        elif mask_type == 'boxes':
            for bb in self.zmask_meta:
                sl = bb['slice']
                zi_st[sl] = 1

        return zi_st

    def classify_by(self, channel, object_classification_model: InstanceSegmentationModel, ):
        # adds a column to self._df
        # do this on a patch basis, i.e. only one object per frame
        obmap_patches = object_classification_model.label_instance_class(
            self.get_raw_patches(channel),  # TODO: enforce df index
            self.get_patch_masks()
        )

        lamap = self.object_id_labels
        om = np.zeros(lamap.shape, dtype=lamap.dtype)
        # self.df['instance_class'] = np.nan

        df = self.get_df()
        idx = df.index
        se = pd.Series(data=np.nan, index=idx)

        # assign labels to object map:
        for i in idx:
            # object_id = self.zmask_meta[i]['info'].label
            object_id = df.loc[i, 'label']
            result_patch = mask_largest_object(obmap_patches.iat(i))
            object_class = np.unique(result_patch)[1]
            om[self.object_id_labels == object_id] = object_class
            se.loc[i] = object_class

        self.object_class_map = InMemoryDataAccessor(om)

    def run_exports(self, where, channel, prefix, params: RoiSetExportParams):
        if not self.count:
            return
        raw_ch = self.acc_raw.get_one_channel_data(channel)
        for k in params.dict().keys():
            subdir = where / k
            pr = prefix
            kp = params.dict()[k]
            if kp is None:
                continue
            if k == 'patches_3d':
                files = export_patches_from_zstack(
                    subdir, raw_ch, self.zmask_meta, prefix=pr, make_3d=True, **kp
                )
            if k == 'patches_2d_for_annotation':
                files = export_multichannel_patches_from_zstack(
                    subdir, self.acc_raw, self.zmask_meta, prefix=pr, make_3d=False, ch_white=channel,
                    bounding_box_channel=1, bounding_box_linewidth=2, **kp,
                )
            if k == 'patches_2d_for_training':
                files = export_multichannel_patches_from_zstack(
                    subdir, self.acc_raw, self.zmask_meta, ch_white=channel, prefix=pr, make_3d=False, **kp
                )
                df_patches = pd.DataFrame(files)
                self.df = pd.merge(self.df, df_patches, left_index=True, right_on='df_index').drop(columns='df_index')
                self.df['patch_id'] = self.df.apply(lambda _: uuid4(), axis=1)
            if k == 'patch_masks':
                self.export_patch_masks(subdir, prefix=pr, **params.patch_masks)
            if k == 'annotated_zstacks':
                annotated = InMemoryDataAccessor(
                    draw_boxes_on_3d_image(raw_ch.data, self.zmask_meta, **kp)
                )
                write_accessor_data_to_file(subdir / (pr + '.tif'), annotated)
            if k == 'object_classes':
                write_accessor_data_to_file(subdir / (pr + '.tif'), self.object_class_map)
            if k == 'dataframe':
                dfpa = subdir / (pr + '.csv')
                dfpa.parent.mkdir(parents=True, exist_ok=True)
                self.df.to_csv(dfpa, index=False)


def project_stack_from_focal_points(
        xx: np.ndarray,
        yy: np.ndarray,
        zz: np.ndarray,
        stack: GenericImageDataAccessor,
        degree: int = 2,
) -> np.ndarray:
    """
    Given a set of 3D points, project a multichannel z-stack based on a surface fit of the provided points
    :param xx: vector of point x-coordinates
    :param yy: vector of point y-coordinates
    :param zz: vector of point z-coordinates
    :param stack: z-stack to project
    :param degree: order of polynomial to fit
    :return: multichannel 2d projected image array
    """
    assert xx.shape == yy.shape
    assert xx.shape == zz.shape

    poly = PolynomialFeatures(degree=degree)
    X = np.stack([xx, yy]).T
    features = poly.fit_transform(X, zz)
    model = LinearRegression(fit_intercept=False)
    model.fit(features, zz)

    xy_indices = np.indices(stack.hw).reshape(2, -1).T
    xy_features = np.dot(
        poly.fit_transform(xy_indices, zz),
        model.coef_
    )
    zi_image = xy_features.reshape(
        stack.hw
    ).round().clip(
        0, (stack.nz - 1)
    ).astype('uint16')

    return np.take_along_axis(
        stack.data,
        np.repeat(
            np.expand_dims(zi_image, (2, 3)),
            stack.chroma,
            axis=2
        ),
        axis=3
    )
